##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
# Some parts Copyright 2005 Canonical Ltd.
#
##############################################################################
"""
Zope-derived Batching Support

*Do not use this. It is now an internal implementation detail.*
The intent is to get rid of this in favor of moving to extending z3c.batching.
"""


from zope.interface import implements
from zope.interface.common.sequence import IFiniteSequence
from interfaces import IBatch

from zope.cachedescriptors.property import Lazy

# The base batch size, which can be overridden by users of _Batch such
# as BatchNavigator. In Launchpad, we override it via a config option.
BATCH_SIZE = 50

class _Batch(object):
    implements(IBatch)

    def _get_length(self, results):
        # If this batch's contents is smaller than the batch size, it is the
        # last batch and we can deduce the underlying data's size.
        if (self._sliced_list is not None
                and len(self._sliced_list) <= self.size):
            return self.start + len(self._sliced_list)
        # LBYL: we don't want to mask exceptions that len might raise, and we
        # don't have chained exceptions yet.
        if getattr(results, '__len__', None) is None:
            results = IFiniteSequence(results)
        return len(results)

    @property
    def listlength(self):
        if self._listlength is None:
            self._listlength = self._get_length(self.list)
        return self._listlength

    @Lazy
    def trueSize(self):
        """Return the actual size of this batch."""
        # If we have the actual batch data, determine the size
        # from that.
        if self._sliced_list is not None:
            if len(self._sliced_list) >= self.size:
                # This batch is full (and there might be another batch
                # afterwards). Return .size
                return self.size
            else:
                # This batch is not full. Return its (user-visible)
                # length.
                return len(self._sliced_list)

        # Otherwise, calculate the size of this batch using the length
        # of the entire list, which 1) might be cached, and 2) might
        # be cheaper to fetch than a batch of actual data.
        if self.start >= self.listlength:
            return 0
        elif self.start + self.size >= self.listlength:
            return self.listlength - self.start
        return self.size

    def __init__(self, results, start=0, size=None, _listlength=None):
        if results is None:
            results = []
        self.list = results

        if _listlength is not None:
            self._listlength = _listlength
        else:
            self._listlength = None
        self._sliced_list = None

        if size is None:
            size = BATCH_SIZE
        self.size = size

        self.start = start
        self.end = start + size

    def __len__(self):
        assert self.trueSize >= 0, ('The number of items in a batch should '
            'never be negative.')
        return self.trueSize

    def __getitem__(self, key):
        if key >= self.trueSize:
            raise IndexError, 'batch index out of range'
        # When self.start is negative (IOW, when we are batching over an
        # empty list) we need to raise IndexError here; otherwise, the
        # attempt to slice below will, on objects which don't implement
        # __len__, raise a mysterious exception directly from python.
        if self.start < 0:
            raise IndexError, 'batch is empty'
        return self.list[self.start+key]

    @Lazy
    def sliced_list(self):
        # We use Lazy here to avoid self.__iter__ giving us new objects every
        # time; in certain cases (such as when database access is necessary)
        # this can be expensive.
        #
        # We get one more item than we need, so that we can detect
        # cases where the underlying list ends on a batch boundary
        # without having to check the total size of the list (another
        # expensive operation).
        sliced = list(self.list[self.start:self.end+1])

        # This is the first time we get an inkling of (approximately)
        # how many items are in the list. This is the appropriate time
        # to handle edge cases.
        if len(sliced) == 0:
            # The list is empty. Set .start to an out-of-range value.
            self.start = -1
        self._sliced_list = sliced
        return sliced

    def __iter__(self):
        sliced = self.sliced_list
        if len(sliced) > self.size:
            # The slice contains more than a full batch, indicating
            # that there is another batch beyond this one. But we
            # don't actually want to return the item from the next
            # batch.
            sliced = sliced[:-1]
        else:
            # The slice contains a full batch or less, indicating that
            # there is no batch beyond this one. In this case we want
            # to return the full slice.
            pass
        return iter(sliced)

    def first(self):
        return self[0]

    def last(self):
        return self[self.trueSize-1]

    def __contains__(self, item):
        return item in self.__iter__()

    def nextBatch(self):
        start = self.start + self.size

        # self._listlength is the definitive indicator of whether
        # there is a next batch to be found.
        if self._listlength is not None and start >= self._listlength:
            return None

        # But self._listlength is expensive to calculate, so if it's
        # not already calculated we don't want to use it. Instead, we'll
        # see if there are fewer items in this batch than we'd expect.
        # That's a sign that we're at the end of the list.
        #
        # self.sliced_list tries to return one more object than the
        # batch size. If it returns the batch size, or fewer, then
        # this batch encompasses the end of the list and there is no
        # batch.
        if len(self.sliced_list) <= self.size:
            return None

        return _Batch(
            self.list, start, self.size, _listlength=self._listlength)

    def prevBatch(self):
        # The only case in which we should /not/ offer a previous batch
        # is when we are already at position zero, which also happens
        # when the list is empty.
        if self.start <= 0:
            return None
        start = self.start - self.size
        if start < 0:
            # This situation happens, for instance, when you have a
            # 20-item batch and you manually set your start to 15;
            # in this case, hopping back one batch would be starting at
            # -5, which doesn't really make sense.
            start = 0
        return _Batch(self.list, start, self.size,
                      _listlength=self._listlength)

    def firstBatch(self):
        return _Batch(self.list, 0, size=self.size,
                      _listlength=self._listlength)

    def lastBatch(self):
        # Return the last possible batch for this dataset, at the
        # correct offset.
        last_index = self.listlength - 1
        last_batch_start = last_index - (last_index % self.size)
        if last_batch_start < 0:
            last_batch_start = 0
        return _Batch(self.list, last_batch_start, size=self.size,
                      _listlength=self.listlength)

    def total(self):
        return self.listlength

    def startNumber(self):
        return self.start+1

    def endNumber(self):
        # If this batch is completely empty, it makes no sense to ask for an
        # "endNumber" so return None.
        if not len(self.sliced_list):
            return None
        return self.start + min(self.size, len(self.sliced_list))
